#! /bin/csh -f
#
#  Create a coverage of integration times per pointing
#  Needs NEMO, but perhaps varmap(s) can also make this map
#  Needs a special version of uvindex, called uvindex2
#
#  See also the movie maker script, which uses this script
# 
#  march-2011    Created for N1333 project              Peter Teuben
#  9-mar-2011    added to MIS
# 10-may-2012    various updates for CLASSy
# 10-may-2012    change to key=val,                     Marc Pound
#                compute datarange on the fly,
#                allow override of x/y range.
#
# TODO:
#   - check if varmap(s) can do the same without using NEMO
#   - option for uvindex to avoid the patched uvindex2 version
#     or add select= to uvindex?

if ($#argv == 0) goto Usage


if ($?MIR == 0) then
    echo No MIRIAD, trying to load it via astromake
    source /astromake/astromake_start
    astroload miriad
endif

if ($?NEMO == 0) then
    echo No NEMO, trying to load it via astromake
    source /astromake/astromake_start
    astroload nemo
endif

#   some defaults that could be changed, note the astro convention in xrange
set xrange="32:-32"
set yrange="-32:32"
set colormap=3

#  command line overrides (just a cute CSH trick)
foreach _arg ($*)
  set $_arg
end

if ( $?srcname == 0 ) then
  echo "srcname must be specified"
  goto Usage
endif

if ( $?tab == 0 ) then
  echo "tab file must be specified"
  goto Usage
endif

# must be tricky with parenthesis to allow
# x/yrange to be +N:-M if requested
set xfoo=`echo \($xrange\)|sed s/:/")-("/`
set yfoo=`echo \($yrange\)|sed s/:/")-("/`
set nx=`calc "abs(4*($xfoo))"`
set ny=`calc "abs(4*($yfoo))"`

#echo $nx,$ny
#echo SRC: $srcname
#echo TAB file: $tab

# to create an (dra,ddex,inttime) need to use uvindex2, special clone of uvindex

# e.g.
# uvindex2 vis=23feb.mir | grep ^SVS | awk '{print $4,$5,$6}' > 23feb.pos


# convert all the pos files to a NEMO snapshot
# note the dra,ddec are in arcsec, and we like the numbers in arcmin

rm -rf $srcname.snap $srcname.fits $srcname.mir

# gather the tables, but convert arcsec to arcmin
cat $tab | tabmath - - %1/60,%2/60,%3 all > junk.tab
tabtos junk.tab $srcname.snap "" x,y,m 
#  the -8:8, 64x64 grid and %1/16 are all related to get them to come out this way
snapgrid $srcname.snap - all $xrange $yrange nx=$nx ny=$ny evar=m |\
  ccdmath - - '%1/16' |\
  ccdfits - $srcname.fits
fits in=$srcname.fits out=$srcname.mir op=xyin
#   mask all 0's
# puthd in=$srcname.mir/datamin value=0 type=real
#minmax in=$srcname.mir
immask in=$srcname.mir options=datamin flag=false
minmax in=$srcname.mir

if ( $?datamin == 0 ) then 
    set datamin=`itemize in=$srcname.mir/datamin|tail -1 |awk '{print $NF}'`
endif
if ( $?datamax == 0 ) then 
    set datamax=`itemize in=$srcname.mir/datamax|tail -1 |awk '{print $NF}'`
endif

#   plot
cgdisp in=$srcname.mir \
    range=$datamin,$datamax,lin,$colormap labtyp=absnat options=full,wedge \
    device=${srcname}_inttime.gif/gif 
cgdisp in=$srcname.mir \
    range=$datamin,$datamax,lin,$colormap labtyp=absnat options=full,wedge \
    device=${srcname}_inttime.ps/ps
histo in=$srcname.mir

exit 0

Usage:
  echo "Usage: $0 srcname= tab= [xrange=min:max] [yrange=min:max] [datamin=] [datamax=] [colormap=]"
  exit 1
