#!/usr/bin/python
# Script to make a movie of uvcoverage as a function of position
# Note: vis file should be made an argument rather than hardcoded
# $Id: uvcoverage.py,v 1.4 2012/04/28 19:04:16 mpound Exp $

import sys, math, parfile, os

class MovieMaker(object) :
    def __init__(self,vis,source,line,sort) :
        self._source  = source
        self._vis = vis
        self._mosaicFile = "/tmp/mosaic.%s" % (source)
        self._line = line
        self._sort = sort

    def create(self) :
        """Make the movie"""
        cmd = "/bin/rm -rf /tmp/moviemakerjunk.* %s" % (self._mosaicFile)
        print cmd
        os.system(cmd)

        cmd = "uvcat vis=%s select=source'(%s)' out=/tmp/moviemakerjunk.vis > /dev/null" % ( self._vis, self._source )
        print cmd
        os.system(cmd)

        cmd = "uvlist vis=/tmp/moviemakerjunk.vis options=brief,list recnum=0 log=/tmp/moviemaker.log > /dev/null"
        print cmd
        os.system(cmd)

        cmd = "grep %s /tmp/moviemaker.log |awk '{print $(NF-1)\" \"$NF}' |sort -u > %s " % ( self._source, self._mosaicFile )
        print cmd
        os.system(cmd)

        f=open(self._mosaicFile,"r")
        lines = f.read().splitlines();
        f.close();
        #print lines
        nframes = len(lines);
        i = 1
        print "Creating %d movie frames with smauvplt and convert..." % (nframes)
#        if ( sort <> None ) :
#            lines.sort()
#            print lines

        for line in lines:
          s="%d.." % i
          sys.stdout.write(s)
          sys.stdout.flush()
          [dra,ddec] = line.split()
          ddeclow = float(ddec) - 5.0
          ddechigh = ddeclow + 10.0
          dralow = float(dra) - 5.0
          drahigh = dralow + 10.0
          cmd = "smauvplt vis=%s line=%s options=notitle,nobase,equal,zero axis=uc,vc select=source\"(%s),dra(%f,%f),ddec(%f,%f)\" xrange=-30,30 yrange=-30,30 dotsize=5 device=uv%05d.gif/gif > /dev/null" % ( self._vis, self._line, self._source, dralow, drahigh, ddeclow, ddechigh, i )
          os.system( cmd )

          cmd ="convert -gravity Northeast -font helvetica -pointsize 36 -draw \"text 350,100 '(%s,%s)'\" uv%05d.gif movieUV%05d.jpg" % ( dra, ddec, i, i )
          os.system( cmd )
          i=i+1

        #print "/n/lupus2/mpound/installed/bin/gifsicle xuv*.gif > movieUV.gif"
        # Can't go straight from gif to avi for some reason, never processes more
        # than first frame.
        # Note ffmpeg understands sprintf format!
        #print "convert movieUV.gif movieUV%05d.jpg"
        print "Compiling %d movie frames into movieUV.avi ..." % (nframes)
        cmd = "ffmpeg -r 6 -i movieUV%05d.jpg -y -an movieUV.avi > /dev/null"
        print cmd
        os.system(cmd)
        cmd = "/bin/rm -rf movieUV*.jpg xuv*.gif uv*.gif"
        print cmd
        os.system(cmd)
        print "DONE"

def usage() :
    print "This script makes an AVI movie of uv-coverage as a function"
    print "of mosaic position. You should uvcat all your nightly vis files"
    print "together into one file first."
    print " "
    print "\033[1mUsage: uvcoverage.py vis=<visfile> source=<sourcename> [line=]\033[0m"
    print "line is the MIRIAD line keyword. Default: wide,1"
    

if __name__ == '__main__':
    p = parfile.ParFile()
    p.set('vis',None)
    p.set('source',None)
    p.set('line',"wide,1")
    p.set('sort',None)
    p.argv(sys.argv)

    vis     = p.get('vis')
    source  = p.get('source')
    line    = p.get('line')
    sort    = p.get('sort')

#    print "vis = :%s:" % vis 
#    print "source = :%s:" % source
#    print "line = :%s:" % source
    if ( vis == None ) :
       print "#### vis must be specified ####"
    if ( source == None ) :
       print "#### source must be specified ####"
    if ( source == None or vis == None ) :
       usage()
    else :
       mm = MovieMaker(vis,source,line, sort)
       mm.create()

