#! /bin/csh -f
#
#
#     CLASSy automated data reduction pipeline
#
#     this can run as crontab as well
#
# Some sample CPU timings as the project and scripts grew in 2012....
#  1-jun    1:15 for all D array with UV and no SD, only UV continuum mapping working
#
# @todo
#   - antpos
#   - bunflag, does it help?
#   - SD
#   - UV line
#   - borrowing bandpass

#   commandline parameters
set today = `date  +%a`
#   this will become $MIS/cat/n1333.lis
unset dirs
#   if you set a pipefile, it will use this instead of the one liners
#   must be an absolute reference, or must live in $MIS/cat
#   args.... @todo in pipe we can't yet pass on parameters that are not pipepars
unset pipefile


#  get MIRIAD if needed
if ($?MIR == 0) then
  source /astromake/astromake_start
  astroload miriad
endif

#  get NEMO if needed
if ($?NEMO == 0) then
  source /astromake/astromake_start
  astroload nemo
endif

# do the work in a fresh directory (using today=test you could override this)

rm -rf Nightly_$today
mkdir  Nightly_$today
rm -f  Nightly_latest
ln -s  Nightly_$today Nightly_latest
cd     Nightly_$today

# get a fresh MIS?

if ($?MIS == 0) then
  cvs -Q co mis
  pushd mis
  ./configure
  source mis_start.csh
  popd
endif

# use the following projects (using dirs= you could override this, use absolute reference)
set dirs=$MIS/cat/classy.lis 


#    now, finally before we start, parse command line arguments
foreach arg ($*)
  set $arg
end

#  copy the current dirs
if (! -e classy.lis) cp $MIS/cat/classy.lis .

#  must ensure a default Nightly link exists for scripts such as do_mos1
if (! -e $MIS/Nightly) then
  echo Creating symlink $MIS/Nightly
  ln -s `pwd` $MIS/Nightly
endif

if (! -e $MIS/rawdata/CLASSy) then
  echo Fixing $MIS/rawdata/CLASSy
  ln -s /chara7nb/teuben/CLASSy $MIS/rawdata/CLASSy
endif

#  start
pipepar -c step=start
 
#  create the empty project directories with the previously pipesave'd MIS files
piperun -c -o pipesetup.log $dirs 'pipepar -c ; pipesetup project=%s'
pipepar step=pipesetup


#  maybe set some parameters to test...
if (0) then
  piperun $dirs pipepar bunflag=1
endif

#  there is no MIS command for this yet, but if you had some new global 
#  parameters in some global mis.def file, this would be the place to
#  apply those, e.g.
#      piperun $dirs "pipepar imsize=128 cell=1 itermax=1000"
#  we might have a new method to shortcut this. 

# get data, make report
piperun -o getdata.log  $dirs 'getdata link=0'
pipepar step=getdata
piperun -o report.log   $dirs report
pipepar step=report

# create separate .lis files for each source we have
set srcs=(`piperun classy.lis pipepar -v srcname | sort | uniq`)
foreach src ($srcs)
  piperun classy.lis pipepar -v project -v srcname |\
       grep -w $src >> classy_$src.lis
end

foreach src ($srcs)
  piperun classy.lis pipepar -v project -v srcname |\
       grep -w $src | awk '{print $1}' > classy_$src.lis
end

if (1) then
  # make coverage maps
  piperun -o inttime.log  $dirs  map_inttime
  pipepar step=map_inttime
  foreach src ($srcs)
    piperun classy_$src.lis cat xyt.tab > $src.pos  
    mk_map_inttime.csh srcname=$src tab=$src.pos >& mk_map_inttime_$src.log
    pipepar step=mk_map_inttime_$src
  end
endif

if (0) then
    # single dish
    piperun -o uvcatSD.log  $dirs do_uvcatSD
    piperun -o reduceSD.log $dirs 'do_reduceSD device=/null sleep=0'
    piperun -o mapSD.log    $dirs 'do_mapSD device=/null'
    # another median to create final SD cubes
    #do_mapSDmedian   >& mapSDmedian.log
    do_mapSDmedian4>& mapSDmedian4.log
endif

if (1) then
    # interferometry
    piperun -o uvcat1.log   $dirs do_uvcat1
    pipepar step=uvcat1
    piperun -o uvcat2.log   $dirs do_uvcat2
    pipepar step=uvcat2
    piperun -o inspect1.log $dirs do_inspect1
    pipepar step=inspect1
    piperun -o cal2.log     $dirs do_cal2
    pipepar step=cal2
    piperun -o cal4.log     $dirs do_cal4 bootstrap=0
    pipepar step=cal4
    if (0) then
      piperun -o map2.log     $dirs do_map2
      pipepar step=map2
    endif
endif

#
set vis=uvdata_wide.pb.cal
foreach src ($srcs)
  set lis=${src}_${vis}.lis
  rm -f $lis
  foreach p (`cat classy_$src.lis`)
    if (-e $p/$vis) then
      echo $p/$vis >> $lis
    endif
  end
  do_map2 vis=@$lis   map=${src}_cont       srcname=$src  cuffoff=0.003
  pipepar step=map2-all
  if (1) then
    # subarray just 10+6m  and   just 3.5m
    do_map2 vis=@$lis   map=${src}_cont_110   srcname=$src  cuffoff=0.003  ants=1,1,0
    do_map2 vis=@$lis   map=${src}_cont_001   srcname=$src  cuffoff=0.003  ants=0,0,1 
  endif
end

# piperun classy.lis pipepar -p %-12s -v srcname -v trial -v rmscal -v rmsres -v rmsres1

if (1) then
  # for the calibrators, make a listing of recent fluxes relevant for us.
  set cals=(`piperun $dirs pipepar -v pcalname | sort | uniq`)
  set ical=0
  foreach cal ($cals)
    @ ical++
    calflux source=$cal  device=calflux_$cal.ps/ps \
	freq=95 delfreq=40 date=11jan01 deldate=1000 yrange=0,40 \
	> calflux_$cal.log
  end
endif

# get the fluxes of a few point like sources in the continuum

set m=SERPM_cont
imfit in=$m/res region='arcsec,box(150,-200,170,-180)' object=gauss  >& $m/a.imfit
imfit in=$m/res region='arcsec,box(95,-110,115,-90)'   object=gauss  >& $m/b.imfit
imfit in=$m/res region='arcsec,box(134,-65,155,-45)'   object=gauss  >& $m/c.imfit
imfit in=$m/res region='arcsec,box(-8,14,12,34)'       object=gauss  >& $m/d.imfit
imfit in=$m/res region='arcsec,box(12,94,32,114)'      object=gauss  >& $m/e.imfit
imfit in=$m/res region='arcsec,box(-32,98,-12,118)'    object=gauss  >& $m/f.imfit
imfit in=$m/res region='arcsec,box(-12,-124,8,144)'    object=gauss  >& $m/g.imfit

set m=B1S_cont
imfit in=$m/res region='arcsec,box(284,216,304,236)'   object=gauss  >& $m/a.imfit
imfit in=$m/res region='arcsec,box(284,234,304,254)'   object=gauss  >& $m/b.imfit
imfit in=$m/res region='arcsec,box(240,343,260,363)'   object=gauss  >& $m/c.imfit

exit 0

# some QA stuff
mk_gaintabs.csh >& mk_gaintabs.log

# combined maps for continuum and 4 channel versions of the 3 molecules
# in 4 different array and antenna selection choices
do_mos1            >& mos1.log
do_mos1 config=D   >& mos1_D.log
do_mos1 config=E   >& mos1_E.log
do_mos1 ants=LOW   >& mos1_LOW.log


# normally at the very end of all the work, you want to save it
# but in a nightly we don't expect to have to do this, since things
# never really (should) change
# these files are now copied to your $MIS/def tree
piperun -o pipesave.log     $dirs pipesave

# if you are inspecting data, you now probably want to commit the 
# mis.def (and other) files back from $MIS/def into the CVS repository 
# with the 'cvs commit' command, or the shortcut in mis:
echo mis -i
